%{

/*
   cfg.y -- PIKT piktc parser for all .cfg files except systems.cfg

   $Header: /var/cvs/PIKT/pikt/src/piktc/cfg.y,v 1.57 2004/01/29 15:06:08 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   ----------------------------------------------------------------------------

   this is the sausage factory; not for the squeamish; enter at your own risk!

*/

#include "../lib/common.h"
#include "piktc.h"

char _cfgparse_[] = "cfgparse";

#define CHKARYNUM	if (arynum >= MAXARY) { \
				logexit(OUTPUT, LOG_ERR, ERRID_ARYNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXARY (%d) exceeded", aidp->name, MAXARY); \
			} \
			else if (arynum < 0) { \
				logexit(OUTPUT, LOG_ERR, ERRID_ARYNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, arynum < 0", aidp->name); \
			}

#define CHKIFNUM	if (ifnum > MAXIFS) \
				logexit(OUTPUT, LOG_ERR, ERRID_IFNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXIFS (%d) exceeded", aidp->name, MAXIFS); \
			else if (ifnum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_IFNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, ifnum < 0", aidp->name)

#define CHKWHILENUM	if (whilenum > MAXWHILES) \
				logexit(OUTPUT, LOG_ERR, ERRID_WHILENUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXWHILES (%d) exceeded", aidp->name, MAXWHILES); \
			else if (whilenum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_WHILENUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, whilenum < 0", aidp->name)

#define CHKREPEATNUM	if (repeatnum > MAXREPEATS) \
				logexit(OUTPUT, LOG_ERR, ERRID_REPEATNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXREPEATS (%d) exceeded", aidp->name, MAXREPEATS); \
			else if (repeatnum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_REPEATNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, repeatnum < 0", aidp->name)

#define CHKFORNUM	if (fornum > MAXFORS) \
				logexit(OUTPUT, LOG_ERR, ERRID_FORNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXFORS (%d) exceeded", aidp->name, MAXFORS); \
			else if (fornum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_FORNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, fornum < 0", aidp->name)

#define CHKSWITCHNUM	if (switchnum > MAXSWITCHS) \
				logexit(OUTPUT, LOG_ERR, ERRID_SWITCHNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXSWITCHS (%d) exceeded", aidp->name, MAXSWITCHS); \
			else if (switchnum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_SWITCHNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, switchnum < 0", aidp->name)

#define CHKLOOPNUM	if (loopnum > MAXLOOPS) \
				logexit(OUTPUT, LOG_ERR, ERRID_LOOPNUM_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, MAXLOOPS (%d) exceeded", aidp->name, MAXLOOPS); \
			else if (loopnum < 0) \
				logexit(OUTPUT, LOG_ERR, ERRID_LOOPNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, loopnum < 0", aidp->name)

#ifdef DEEPDEBUG
#define DUMP(token)	if (isdebug) logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, _cfgparse_, FALSE, "@%s@", token)
#else
#define DUMP(token)	;
#endif

int doalert, doprogram, dofile, doobjects;

int ifnum, whilenum, repeatnum, fornum, switchnum;
int loopnum;

char *mcrnam;
unsigned int mcrnamlen;
unsigned int mcrprmnum = 0;
signed char mpn[] = { '\0', '\0' };
char *mcrnamprm, *mcrnamprmnum;

ident *midp;	/* macro def idents */
ident *mpidp;	/* macro prm idents */
ident *aidp;	/* alarm idents */
ident *tidp;	/* alert idents */
ident *pidp;	/* program idents */
ident *fidp;	/* file idents */
ident *oidp;	/* objects idents */

int i;
int p;

int dim[MAXARY+1];
int arynum;

%}

%union {
	int ival;
	signed char *sval;
}

%token _MCRCFGBOF _ALMCFGBOF _ALTCFGBOF _PRGCFGBOF _FILCFGBOF _OBJCFGBOF
%token <sval> _MACRODEF _MACROPRM
%token <sval> _STANZA _OBJECT _LINE
%token _PATH _MODE _UID _GID
%token _TIMING _DRIFT
%token <sval> _TIMEVAL
%token _PRIORITY _NICECMD _MAILCMD _LPCMD _EXECCMD
%token _ALARMS
%token <sval> _WORD
%token <sval> _STATUS _ACTIVE _INACTIVE
%token <sval> _LEVEL _EMERG _ALERT _CRIT _ERR _WARNING _NOTICE _INFO _DEBUG
%token <sval> _TASK
%token <sval> _INPUT _PROCESS _REGFILE _LOGFILE
%token <sval> _FILTER
%token <sval> _SEPARATORS _DAT _KEYS
%token <sval> _INIT
%token <sval> _BEGIN _END
%token <sval> _RULE
%token <sval> _DO
%token <sval> _IF _ELIF _ELSE _ENDIF
%token <sval> _WHILE _ENDWHILE _REPEAT _UNTIL _FOR _FOREACH _ENDFOR
%token <sval> _IN
%token <sval> _BREAK _CONTINUE
%token <sval> _FORLP _SWITCHLP
%token <sval> _SWITCH _CASE _DEFAULT _BREAKSWITCH _ENDSWITCH
%token <sval> _AGAIN _LEAVE _REDO _NEXT _SKIP _LAST _PAUSE _QUIT _DIE
%token <sval> _OUTPUT _MAIL _PRINT _PIKTLOG _SYSLOG
%token <sval> _SET _UNSET _EXEC _WAIT
%token <sval> _STR _STRVAR
%token <sval> _POSITION _RANGE
%token <sval> _VARNUM _VARSTR _VARHSTSTR _VARHSTNUM _VARPRCSTR _VARPRCNUM
%token <sval> _FLHNDL
%token <sval> _VALNUM _VALSTR
%token <sval> _NOP _SOP _FOP _AOPNUM _AOPSTR _NREL _SREL
%token <sval> _NUM _INT
%token <sval> _FNNUM _FNSTR
%token <sval> _NOT _AND _OR

%type <sval> macrostanza alarmstanza alertstanza programstanza filestanza objectstanza
%type <sval> position varnum varstr varnow
%type <sval> aopnum aopstr
%type <ival> exp ndx
%type <ival> ndxnums

%left _OR
%left _AND
%nonassoc _NREL _SREL
%left '+' '-' _NOP _SOP
%nonassoc _FOP
%nonassoc _NOT '~' _UNOP

%expect 4

%%

start:		mcrcfg
		almcfg
		altcfg
		prgcfg
		filcfg
		objcfg
		;

mcrcfg:		/* empty */
		| _MCRCFGBOF
		| _MCRCFGBOF macrodefs
		;

almcfg:		/* empty */
		| _ALMCFGBOF
		| _ALMCFGBOF alarmdefs
		;

altcfg:		/* empty */
		| _ALTCFGBOF
		| _ALTCFGBOF alertdefs
		;

prgcfg:		/* empty */
		| _PRGCFGBOF
		| _PRGCFGBOF programs
		;

filcfg:		/* empty */
		| _FILCFGBOF
		| _FILCFGBOF files
		;

objcfg:		/* empty */
		| _OBJCFGBOF
		| _OBJCFGBOF objectsets
		;

macrodefs:	macrodef
		| macrodefs macrodef
		;

macrodef:	macrostanza {
			midp = regident($1, MCR, FALSE);
			midp->prmnum = 0;
		}
		macrodefcomps
		| macrostanza
		'(' macroprms ')' {
			midp = regident($1, MCR, FALSE);
			midp->prmnum = mcrprmnum;
		}
		macrodefcomps
		| error { YYABORT; }
		;

macrostanza:	_STANZA {
			DUMP($1);
			mcrnam = $1;
			mcrnamlen = strlen($1);
			mcrprmnum = 0;
			$$ = $1;
		}
		;

macroprms:	macroprm
		| macroprms ',' macroprm
		;

macroprm:	_MACROPRM {
			DUMP($1);
			mcrnamprm = mycalloc((mcrnamlen+strlen($1)+2), sizeof(char), _cfgparse_);
			sprintf(mcrnamprm, "%s%c%s", mcrnam, MCRLNK, $1);
			mpidp = regident(mcrnamprm, MCR, FALSE);
			if (mcrprmnum >= MAXMCRPRMS) {
				logreturn(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETERS_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "number of macro parameters exceeds maximum (%d)", MAXMCRPRMS);
				bailout = TRUE;
				mcrprmnum--;
			}
			mpn[0] = ++mcrprmnum;
			regcomponent(mpidp, mystrdup(mpn, _cfgparse_));
			mcrnamprmnum = mycalloc(mcrnamlen+3, sizeof(char), _cfgparse_);
			sprintf(mcrnamprmnum, "%s%c%c", mcrnam, MCRLNK, mcrprmnum);
			mpidp = regident(mcrnamprmnum, MCR, FALSE);
			free($1);
		}
		;

macrodefcomps:	macrodefcomp
		| macrodefcomps macrodefcomp
		;

macrodefcomp:	_MACRODEF {
			if (*$1 && (*$1 != '\n'))
				DUMP($1);
			regcomponent(midp, parametize(mcrnam, mcrnamlen, $1));
			free($1);
		}
		;

alarmdefs:	alarmdef
		| alarmdefs alarmdef
		;

alarmdef:	alarmstanza scriptdef { ; }
		;

scriptdef:	begin
		| end
		| begin end
		| init begin
		| init end
		| init begin end
		| init dataspec begin
		| init dataspec end
		| init dataspec begin end
		| init dataspec rules
		| init dataspec begin rules
		| init dataspec rules end
		| init dataspec begin rules end
		| statements
		;

alarmstanza:	_STANZA {
			DUMP($1);
			aidp = regident($1, ALM, FALSE);
			$$ = $1;
		}
		| error { yyclearin; stzerr = FALSE; }
		_STANZA {
			DUMP($3);
			aidp = regident($3, ALM, FALSE);
			$$ = $3;
		}
		;

init:		_INIT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _INIT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		status level task
		;

status:		_STATUS _ACTIVE {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _STATUS _INACTIVE {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

level:		_LEVEL _EMERG {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _ALERT {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _CRIT {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _ERR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _WARNING {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _NOTICE {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _INFO {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _LEVEL _DEBUG {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

task:		_TASK _STR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

dataspec:	input filter separators inputdata keys
		;

input:		_INPUT _STR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		| _INPUT _PROCESS _STR {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		}
		| _INPUT _REGFILE _STR {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		}
		| _INPUT _REGFILE _STR _STR {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			DUMP($4);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
			regcomponent(aidp, $4);
		}
		| _INPUT _LOGFILE _STR {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		}
		;

filter:		/* empty */ { ; }
		| _FILTER _STR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

separators:	/* empty */ { ; }
		| _SEPARATORS _STR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

inputdata:	/* empty */
		| data
		;

data:		datum
		| data datum
		;

datum:		_DAT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		varnowset position {
			DUMP($4);
			regcomponent(aidp, $4);
		}
		| _DAT _STR {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		;

position:	_POSITION { $$ = $1; }
		| _RANGE { $$ = $1; }
		;

keys:		/* empty */ { ;	}
		| _KEYS {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		keyset
		;

keyset:		varnowset { ; }
		| keyset varnowset { ; }
		;

varnowset:	varnow { ; }
		| varnow '[' _INT ']' {
			DUMP("[");
			DUMP($3);
			DUMP("]");
			regcomponent(aidp, mystrdup("[", _cfgparse_));
			regcomponent(aidp, $3);
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		| varnow '[' _STR ']' {
			DUMP("[");
			DUMP($3);
			DUMP("]");
			regcomponent(aidp, mystrdup("[", _cfgparse_));
			regcomponent(aidp, $3);
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		;

varnow:		_VARNUM {
			DUMP($1);
			regcomponent(aidp, $1);
			$$ = $1;
		}
		| _VARSTR {
			DUMP($1);
			regcomponent(aidp, $1);
			$$ = $1;
		}
		;

begin:		_BEGIN {
			ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		statements
		;

end:		_END {
			ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		statements
		;

rules:		rule
		| rules rule
		;

rule:		_RULE {
			ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		statements
		;

statements:	statement
		| statements statement
		;

statement:	do
		| ifblk
		| whileblk
		| repeatblk
		| forblk
		| switchblk
		| break
		| continue
		| again
		| leave
		| redo
		| next
		| skip
		| last
		| pause
		| quit
		| die
		| output
		| set
		| unset
		| exec
		;

do:		_DO {
			DUMP($1);
			regcomponent(aidp, $1);
		} exp
		;

ifblk:		if endif { ifnum--; CHKIFNUM; }
		| if statements endif { ifnum--; CHKIFNUM; }
		| if else endif { ifnum--; CHKIFNUM; }
		| if statements else endif { ifnum--; CHKIFNUM; }
		| if else statements endif { ifnum--; CHKIFNUM; }
		| if statements else statements endif { ifnum--; CHKIFNUM; }
		| if elifs endif { ifnum--; CHKIFNUM; }
		| if statements elifs endif { ifnum--; CHKIFNUM; }
		| if elifs else endif { ifnum--; CHKIFNUM; }
		| if statements elifs else endif { ifnum--; CHKIFNUM; }
		| if elifs else statements endif { ifnum--; CHKIFNUM; }
		| if statements elifs else statements endif { ifnum--; CHKIFNUM; }
		;

elifs:		elif
		| elif statements
		| elifs elif
		| elifs elif statements
		;

if:		_IF {
			ifnum++;
			CHKIFNUM;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		conditions
		;

elif:		_ELIF {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		conditions
		;

else:		_ELSE {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

endif:		_ENDIF {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

whileblk:	while endwhile { whilenum--; loopnum--; }
		| while statements endwhile { whilenum--; loopnum--; }
		;

while:		_WHILE {
			whilenum++;
			loopnum++;
			CHKWHILENUM;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		conditions
		;

endwhile:	_ENDWHILE {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

repeatblk:	repeat until { repeatnum--; loopnum--; }
		| repeat statements until { repeatnum--; loopnum--; }
		;

repeat:		_REPEAT {
			repeatnum++;
			loopnum++;
			CHKREPEATNUM;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

until:		_UNTIL {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		conditions
		;

forblk:		for endfor { fornum--; loopnum--; }
		| for statements endfor { fornum--; loopnum--; }
		| foreach endfor { fornum--; loopnum--; }
		| foreach statements endfor { fornum--; loopnum--; }
		;

for:		_FOR {
			DUMP($1);
			regcomponent(aidp, $1);
		} forexps
		| _FORLP {
			DUMP($1);
			free($1);
			regcomponent(aidp, mystrdup("for (", _cfgparse_));
		} forexps ')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		;

forexps:	_VARNUM '=' {
			fornum++;
			loopnum++;
			CHKFORNUM;
			DUMP($1);
			DUMP("=");
			regcomponent(aidp, $1);
			regcomponent(aidp, mystrdup("=", _cfgparse_));
		} expnum { ; }
		conditions
		_VARNUM aopnum {
			DUMP($7);
			DUMP($8);
			regcomponent(aidp, $7);
			regcomponent(aidp, $8);
		} expnum
		;

foreach:	_FOREACH _FNNUM '(' _VARSTR ',' {
			fornum++;
			loopnum++;
			CHKFORNUM;
			DUMP($1);
			DUMP($2);
			DUMP($4);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, mystrdup("(", _cfgparse_));
			regcomponent(aidp, $4);
			regcomponent(aidp, mystrdup(",", _cfgparse_));
		}
		args ')' {
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		| _FOR _VARSTR _IN _FNNUM '(' {
			fornum++;
			loopnum++;
			CHKFORNUM;
			DUMP($2);
			DUMP($4);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
			regcomponent(aidp, $4);
			regcomponent(aidp, mystrdup("(", _cfgparse_));
		}
		args ')' {
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		;

endfor:		_ENDFOR {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

switchblk:	switch cases endswitch { switchnum--; }
		| switch caseblks cases endswitch { switchnum--; }
		| switch caseblks endswitch { switchnum--; }
		| switch caseblks defaultblk endswitch { switchnum--; }
		| switch defaultblk endswitch { switchnum--; }
		;

caseblks:	caseblk
		| caseblks caseblk
		;

caseblk:	cases statements
		| cases breakswitch
		| cases statements breakswitch
		;

defaultblk:	default
		| default statements
		| default breakswitch
		| default statements breakswitch
		| cases default
		| cases default statements
		| cases default breakswitch
		| cases default statements breakswitch
		;

cases:		case
		| cases case
		;

switch:		_SWITCH {
			DUMP($1);
			regcomponent(aidp, $1);
		}switchexp
		| _SWITCHLP {
			DUMP($1);
			free($1);
			regcomponent(aidp, mystrdup("switch (", _cfgparse_));
		} switchexp ')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		;

switchexp:	{
			switchnum++;
			CHKSWITCHNUM;
		} exp
		;

case:		_CASE {
			DUMP($1);
			regcomponent(aidp, $1);
		} exp
		;

default:	_DEFAULT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

breakswitch:	_BREAKSWITCH {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

endswitch:	_ENDSWITCH {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

break:		_BREAK {
			if (loopnum == 0)
				logreturn(OUTPUT, LOG_ERR, ERRID_BREAK_WITHOUT_LOOP, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, break with no corresponding loop", aidp->name);
			else
				CHKLOOPNUM;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

continue:	_CONTINUE {
			if (loopnum == 0)
				logreturn(OUTPUT, LOG_ERR, ERRID_CONTINUE_WITHOUT_LOOP, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, continue with no corresponding loop", aidp->name);
			else
				CHKLOOPNUM;
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

again:		_AGAIN {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

leave:		_LEAVE {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

redo:		_REDO {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

next:		_NEXT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

skip:		_SKIP {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		expnum
		;

last:		_LAST {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

pause:		_PAUSE {
			DUMP($1);
			regcomponent(aidp, $1);
		} expnum
		;

quit:		_QUIT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

die:		_DIE {
			DUMP($1);
			regcomponent(aidp, $1);
		} expstr
		;

output:		_OUTPUT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		expstr
		| _OUTPUT _MAIL {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		}
		expstr
		| _OUTPUT _PRINT {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		} expstr
		| _OUTPUT _LOGFILE _STR {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		} expstr
		| _OUTPUT _PIKTLOG {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		} expstr
		| _OUTPUT _SYSLOG {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		} expstr
		;

set:		_SET _VARNUM aopnum {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		} conditions
		| _SET _VARSTR aopstr {
			DUMP($1);
			DUMP($2);
			DUMP($3);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, $3);
		} expstr
		| _SET _VARNUM '[' {
			DUMP($1);
			DUMP($2);
			DUMP("[");
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, mystrdup("[", _cfgparse_));
			arynum = 0;
		}
		ndx
		']' aopnum {
			DUMP("]");
			DUMP($7);
			regcomponent(aidp, mystrdup("]", _cfgparse_));
			regcomponent(aidp, $7);
		} conditions
		| _SET _VARSTR '[' {
			DUMP($1);
			DUMP($2);
			DUMP("[");
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
			regcomponent(aidp, mystrdup("[", _cfgparse_));
			arynum = 0;
		}
		ndx
		']' aopstr {
			DUMP("]");
			DUMP($7);
			regcomponent(aidp, mystrdup("]", _cfgparse_));
			regcomponent(aidp, $7);
		} expstr
		;

unset:		_UNSET {
			DUMP($1);
			regcomponent(aidp, $1);
		} varnow '[' {
			DUMP("[");
			regcomponent(aidp, mystrdup("[", _cfgparse_));
		}
		expstr
		']' {
			DUMP("]");
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		;

aopnum:		'=' { $$ = mystrdup("=", _cfgparse_); }
		| _AOPNUM { $$ = $1; }
		;

aopstr:		'=' { $$ = mystrdup("=", _cfgparse_); }
		| _AOPSTR { $$ = $1; }
		;

exec:		_EXEC {
			DUMP($1);
			regcomponent(aidp, $1);
		} expstr
		| _EXEC _WAIT {
			DUMP($1);
			DUMP($2);
			regcomponent(aidp, $1);
			regcomponent(aidp, $2);
		} expstr
		;

varstr:		_VARSTR {
			$$ = $1;
		}
		| _VARHSTSTR {
			$$ = $1;
		}
		| _VARPRCSTR {
			$$ = $1;
		}
		;

varnum:		_VARNUM {
			$$ = $1;
		}
		| _VARHSTNUM {
			$$ = $1;
		}
		| _VARPRCNUM {
			$$ = $1;
		}
		;

args:		/* empty */ { ; }
		| arg
		| argcoms arg
		;

argcoms:	argcom
		| argcoms argcom
		;

argcom:		arg
		',' {
			regcomponent(aidp, mystrdup(",", _cfgparse_));
		}
		;

arg:		conditions
		| _FLHNDL {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		;

fnnum:		_FNNUM {
			regcomponent(aidp, $1);
		}
		'(' {
			DUMP("(");
			regcomponent(aidp, mystrdup("(", _cfgparse_));
		}
		args
		')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		;

fnstr:		_FNSTR {
			regcomponent(aidp, $1);
		}
		'(' {
			DUMP("(");
			regcomponent(aidp, mystrdup("(", _cfgparse_));
		}
		args
		')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		;

ndx:		expstr {
			$$ = 1<<1 | STRING;	/*  1 dimension */
		}
		| {
			CHKARYNUM;
			if (arynum <= MAXARY-1)
				dim[arynum++] = 0;
		} ndxnums {
			arynum--;
			CHKARYNUM;
			$$ = $2<<1 | NUMBER;	/* $2 dimensions */
		}
		;

ndxnums:	expnum {
			$$ = ++dim[arynum-1];
		}
		| ndxnums ',' {
			regcomponent(aidp, mystrdup(",", _cfgparse_));
		} expnum {
			$$ = ++dim[arynum-1];
		}
		;

exp:		expstr { $$ = STRING; }
		| expnum { $$ = NUMBER; }
		;

expstr:		_STR {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _STRVAR {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _VALSTR {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _VALSTR '[' {
			DUMP($1);
			DUMP("[");
			regcomponent(aidp, $1);
			regcomponent(aidp, mystrdup("[", _cfgparse_));
		}
		expnum
		']' {
			DUMP("]");
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		| varstr {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| varstr '[' {
			DUMP($1);
			DUMP("[");
			regcomponent(aidp, $1);
			regcomponent(aidp, mystrdup("[", _cfgparse_));
		}
		ndx
		']' {
			DUMP("]");
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		| fnstr
		| expstr
		_SOP {
			DUMP($2);
			regcomponent(aidp, $2);
		}
		expstr
		;
 
expnum:		_NUM {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _INT {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| _VALNUM {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| varnum {
			DUMP($1);
			regcomponent(aidp, $1);
		}
		| varnum '[' {
			DUMP($1);
			DUMP("[");
			regcomponent(aidp, $1);
			regcomponent(aidp, mystrdup("[", _cfgparse_));
		}
		ndx
		']' {
			DUMP("]");
			regcomponent(aidp, mystrdup("]", _cfgparse_));
		}
		| fnnum
		| '(' {
			DUMP("(");
			regcomponent(aidp, mystrdup("(", _cfgparse_));
		} expnum ')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		| '+' {
			DUMP("+");
			regcomponent(aidp, mystrdup("+", _cfgparse_));
		}
		expnum %prec _UNOP
		| '-' {
			DUMP("-");
			regcomponent(aidp, mystrdup("-", _cfgparse_));
		}
		expnum %prec _UNOP
		| '~' {
			DUMP("~");
			regcomponent(aidp, mystrdup("~", _cfgparse_));
		}
		expnum
		| expnum
		_NOP {
			DUMP($2);
			regcomponent(aidp, $2);
		}
		expnum
		| expnum
		'+' {
			DUMP("+");
			regcomponent(aidp, mystrdup("+", _cfgparse_));
		}
		expnum
		| expnum
		'-' {
			DUMP("-");
			regcomponent(aidp, mystrdup("-", _cfgparse_));
		}
		expnum
		;

conditions:	expression
		| relation
		| '[' {
			DUMP("(");
			regcomponent(aidp, mystrdup("(", _cfgparse_));
		} conditions ')' {
			DUMP(")");
			regcomponent(aidp, mystrdup(")", _cfgparse_));
		}
		| _NOT {
			regcomponent(aidp, $1);
		} conditions
		| conditions _AND {
			regcomponent(aidp, $2);
		} conditions
		| conditions _OR {
			regcomponent(aidp, $2);
		} conditions
		;

expression:	expnum
		| _FOP {
			DUMP($1);
			regcomponent(aidp, $1);
		} expstr
		| expstr
		;

relation:	expnum
		_NREL {
			DUMP($2);
			regcomponent(aidp, $2);
		}
		expnum
		| expstr
		_SREL {
			DUMP($2);
			regcomponent(aidp, $2);
		}
		expstr
		;

alertdefs:	alertdef
		| alertdefs alertdef
		;

alertdef:	alertstanza
		timings
		priority
		mailing
		printing
		_ALARMS {
			if (doalert && (piktdconf_toolong($1, timings)) && ! (iscksum || isdiff)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_PIKTDCONF_ARG_LEN_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "for alert %s, in piktd.conf, argument lengths too long", $1);
				doalert = FALSE;
			}
			if (doalert)
				beginalarms($1, timings);
		}
		alarms {
			if (doalert)
				endalarms($1);
			for (i=1; i<=timings; i++) FREEVOID(mins[i]);
			for (i=1; i<=timings; i++) FREEVOID(hrs[i]);
			for (i=1; i<=timings; i++) FREEVOID(doms[i]);
			for (i=1; i<=timings; i++) FREEVOID(mons[i]);
			for (i=1; i<=timings; i++) FREEVOID(dows[i]);
			for (i=1; i<=timings; i++) FREEVOID(drift[i]);
			FREEVOID(priority);
			FREEVOID(nicecmd);
			FREEVOID(mailcmd);
			FREEVOID(lpcmd);
		}
		| alertstanza
		timings
		priority
		command {
			if (doalert)
				beginalarms($1, timings);
			/* fake an alarms.cfg registration; script will
			   consist of just the (alert) name, followed by
			   'exec "<command>"'
			 */
			aidp = regident(mystrdup($1, _cfgparse_), ALM, FALSE);
			regcomponent(aidp, mystrdup("exec", _cfgparse_));
			regcomponent(aidp, mystrdup(execcmd, _cfgparse_));
			if (isdoublecheck || (doalert && ! (ischeck || islist_elements)))
				writealarm($1);
			if (doalert)
				endalarms($1);
			for (i=1; i<=timings; i++) FREEVOID(mins[i]);
			for (i=1; i<=timings; i++) FREEVOID(hrs[i]);
			for (i=1; i<=timings; i++) FREEVOID(doms[i]);
			for (i=1; i<=timings; i++) FREEVOID(mons[i]);
			for (i=1; i<=timings; i++) FREEVOID(dows[i]);
			for (i=1; i<=timings; i++) FREEVOID(drift[i]);
			FREEVOID(priority);
			FREEVOID(nicecmd);
			FREEVOID(execcmd);
		}
		;

alertstanza:	_STANZA {
			DUMP($1);
			tidp = regident($1, ALT, FALSE);
			doalert = (int) (lookupname(cmdalerts,$1) ||
			                 isdoublecheck ||
			                 ((iscksum || isdiff) && ! cmdalerts && ! cmdprograms && ! cmdfiles && ! cmdobjects)
			                );
			timings = 0;
			priority = (char *) NULL;
			nicecmd = (char *) NULL;
			mailcmd = (char *) NULL;
			lpcmd = (char *) NULL;
			execcmd = (char *) NULL;
			$$ = $1;
		}
		;

timings:	/* empty */ {
			/* PIKTNEVER is #define'd as "0 0 0 0 0 0" */
			timings = 1;
			mins[1]  = mystrdup("0", _cfgparse_);
			hrs[1]   = mystrdup("0", _cfgparse_);
			doms[1]  = mystrdup("0", _cfgparse_);
			mons[1]  = mystrdup("0", _cfgparse_);
			dows[1]  = mystrdup("0", _cfgparse_);
			drift[1] = mystrdup("0", _cfgparse_);
		}
		| _TIMING _TIMEVAL _TIMEVAL _TIMEVAL _TIMEVAL _TIMEVAL _DRIFT _TIMEVAL {
			DUMP($2);
			DUMP($3);
			DUMP($4);
			DUMP($5);
			DUMP($6);
			DUMP($8);
			timings = 1;
			mins[1] = $2;
			hrs[1]  = $3;
			doms[1] = $4;
			mons[1] = $5;
			dows[1] = $6;
			if (   strchr($8, '-')
			    || strchr($8, ',')
			    || strchr($8, '*')
			   ) {
				logreturn(OUTPUT, LOG_ERR, ERRID_DRIFT_INVALID, __FILE__, __LINE__, _cfgparse_, FALSE, "bad drift value %s", _cfgparse_, $8);
				bailout = TRUE;
			}
			drift[1] = $8;
		}
		| _TIMING timingsets
		;

timingsets:	timingset
		| timingsets timingset
		;

timingset:	_TIMEVAL _TIMEVAL _TIMEVAL _TIMEVAL _TIMEVAL {
			if (timings >= MAXTIMINGS) {
				logreturn(OUTPUT, LOG_ERR, ERRID_TIMINGS_MAX_EXCEEDED, __FILE__, __LINE__, _cfgparse_, FALSE, "maximum number of timings (%d) exceeded", MAXTIMINGS);
				bailout = TRUE;
				/* if there are more than MAXTIMINGS listed
				   timing specs, timings is not incremented
				   beyond MAXTIMINGS, so there is no
				   possibility of exceeding the array bounds
				   for mins[], drift[], etc.
				 */
			} else {
				timings++;
			}
			DUMP($1);
			DUMP($2);
			DUMP($3);
			DUMP($4);
			DUMP($5);
			mins[timings] = $1;
			hrs[timings]  = $2;
			doms[timings] = $3;
			mons[timings] = $4;
			dows[timings] = $5;
		} drift
		;

drift:		/* empty */ {
			drift[timings] = mystrdup("0", _cfgparse_);
		}
		| _TIMEVAL {
			DUMP($1);
			if (   strchr($1, '-')
			    || strchr($1, ',')
			    || strchr($1, '*')
			   ) {
				logreturn(OUTPUT, LOG_ERR, ERRID_DRIFT_INVALID, __FILE__, __LINE__, _cfgparse_, FALSE, "bad drift value %s", $1);
				bailout = TRUE;
			}
			drift[timings] = $1;
		}
		;

priority:	/* empty */ { ; }
		| _PRIORITY _TIMEVAL {	/* nice level, actually */
			DUMP($2);
			errno = 0;
			p = atoi($2);
			if (errno) {
				logreturn(OUTPUT, LOG_ERR, ERRID_PRIORITY_INVALID, __FILE__, __LINE__, _cfgparse_, TRUE, "bad priority value %s", $2);
				bailout = TRUE;
				p = 0;
			}
			if ((p > 19) || (p < -20)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_PRIORITY_INVALID, __FILE__, __LINE__, _cfgparse_, FALSE, "bad priority value %s, must be -20 <= prio <= 19", $2);
				bailout = TRUE;
			}
			priority = $2;
			if (p) {
				nicecmd = getnicecmd(priority);
			/* else p == 0,
				hence default priority, hence no nicecmd
			 */
			}
		}
		| _NICECMD _STR {
			DUMP($2);
			nicecmd = $2;
		}
		;

mailing:	/* empty */ { ; }
		| _MAILCMD _STR {
			DUMP($2);
			mailcmd = $2;
		}
		;

printing:	/* empty */ { ; }
		| _LPCMD _STR {
			DUMP($2);
			lpcmd = $2;
		}
		;

command:	_EXECCMD _STR {
			DUMP($2);
			execcmd = $2;
		}
		;

alarms:		alarm
		| alarms alarm
		;

alarm:		_WORD {
			DUMP($1);
			if (! lookupident($1, ALMCFG))
				logreturn(OUTPUT, LOG_ERR, ERRID_ALARM_NOT_REGISTERED, __FILE__, __LINE__, _cfgparse_, FALSE, "in %s, %s alarm not registered", cfgfilecurr, $1);
			else if (isdoublecheck || (doalert && ! (ischeck || islist_elements)))
				writealarm($1);
			free($1);
		}
		;

programs:	program
		| programs program
		;

program:	programstanza {
			if (doprogram)
				beginfile($1, $1, mystrdup(MODE_PRG_DEFAULT, _cfgparse_), mystrdup(UID_FIL_DEFAULT, _cfgparse_), mystrdup(GID_FIL_DEFAULT, _cfgparse_), &programs);
		}
		lines {
			if (doprogram)
				endfile($1);
		}
		| programstanza _MODE _INT _UID _INT _GID _INT {
			if (doprogram)
				beginfile($1, $1, $3, $5, $7, &programs);
		}
		lines {
			if (doprogram)
				endfile($1);
		}
		| programstanza _PATH _STR _MODE _INT _UID _INT _GID _INT {
			if (doprogram)
				beginfile($1, $3, $5, $7, $9, &programs);
		}
		lines {
			if (doprogram)
				endfile($3);
		}
		;

programstanza:	_STANZA {
			DUMP($1);
			pidp = regident($1, PRG, FALSE);
			dofile = FALSE;
			doprogram = (int) (lookupname(cmdprograms, $1) ||
			                   ((iscksum || isdiff) && ! cmdalerts && ! cmdprograms && ! cmdfiles && ! cmdobjects)
			                  );
			$$ = $1;
		}
		;

files:		file
		| files file
		;

file:		filestanza {
			if (dofile)
				beginfile($1, $1, mystrdup(MODE_FIL_DEFAULT, _cfgparse_), mystrdup(UID_FIL_DEFAULT, _cfgparse_), mystrdup(GID_FIL_DEFAULT, _cfgparse_), &files);
		}
		lines {
			if (dofile)
				endfile($1);
		}
		| filestanza _MODE _INT _UID _INT _GID _INT {
			if (dofile)
				beginfile($1, $1, $3, $5, $7, &files);
		}
		lines {
			if (dofile)
				endfile($1);
		}
		| filestanza _PATH _STR _MODE _INT _UID _INT _GID _INT {
			if (dofile)
				beginfile($1, $3, $5, $7, $9, &files);
		}
		lines {
			if (dofile)
				endfile($3);
		}
		;

filestanza:	_STANZA {
			DUMP($1);
			fidp = regident($1, FIL, FALSE);
			doprogram = FALSE;
			dofile = (int) (lookupname(cmdfiles, $1) ||
			                ((iscksum || isdiff) && ! cmdalerts && ! cmdprograms && ! cmdfiles && ! cmdobjects)
			               );
			$$ = $1;
		}
		;

lines:		line
		| lines line
		;

line:		_LINE {
			if (doprogram || dofile)
				writefile($1);
			free($1);
		}
		;

objectsets:	objectset
		| objectsets objectset
		;

objectset:	objectstanza {
			if (doobjects)
				beginobjects($1);
		}
		objects {
			if (doobjects)
				endobjects($1);
		}
		;

objectstanza:	_STANZA {
			DUMP($1);
			oidp = regident($1, OBJ, FALSE);
			doobjects = (int) (lookupname(cmdobjects,$1) ||
			                   ((iscksum || isdiff) && ! cmdalerts && ! cmdprograms && ! cmdfiles && ! cmdobjects)
			                  );
			$$ = $1;
		}
		;

objects:	object
		| objects object
		;

object:		_OBJECT {
			if (doobjects)
				writeobject($1);
			free($1);
		}
		;

%%
